import logging
from typing import Optional
from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field
from cg_lims.models.arnold.prep.base_step import BaseStep
from cg_lims.objects import BaseAnalyte

LOG = logging.getLogger(__name__)


class NormalizationOfSamplesForSequencingProcessUDFS(BaseModel):
    lot_nr_dilution_buffer_library_normalization: Optional[str] = Field(
        None, alias="Dilution buffer lot no"
    )
    normalized_library_concentration: Optional[float] = Field(
        None, alias="Final Concentration (nM)"
    )
    library_normalization_method: Optional[str] = Field(None, alias="Method document")


class NormalizationForSequencingFields(BaseStep):
    process_udfs: NormalizationOfSamplesForSequencingProcessUDFS

    class Config:
        allow_population_by_field_name = True


def get_normalization_of_samples(
    lims: Lims, sample_id: str, prep_id: str
) -> Optional[NormalizationForSequencingFields]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="CG002 - Normalization of microbial samples for sequencing",
        optional_step=True,
    )
    if not analyte.artifact:
        return None

    return NormalizationForSequencingFields(
        **analyte.base_fields(),
        process_udfs=NormalizationOfSamplesForSequencingProcessUDFS(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="normalization_for_sequencing",
        workflow="Microbial"
    )



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeProcessUDFS(BaseModel):
    lot_nr_truseq_library_preparation_kit: Optional[str] = Field(
        None, alias="Lot no: TruSeq DNA PCR-Free Sample Prep Kit"
    )
    lot_nr_index: str = Field(..., alias="Lot no: Adaptor Plate")
    lot_nr_beads: str = Field(..., alias="Lot no: SP Beads")
    lot_nr_lucigen_library_preparation_kit: Optional[str] = Field(
        None, alias="Lot no: Lucigen prep kit"
    )
    pcr_instrument_incubation: str = Field(..., alias="PCR machine")
    lot_nr_h2o_library_preparation: str = Field(..., alias="Lot no: Nuclease free water")
    lot_nr_resuspension_buffer_library_preparation: str = Field(
        ..., alias="Lot no: Resuspension buffer"
    )
    library_preparation_method: str = Field(..., alias="Method document")
    lot_nr_etoh_library_preparation: str = Field(..., alias="Ethanol lot")


class EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeArtifactUDF(BaseModel):
    finished_library_concentration: Optional[float] = Field(None, alias="Concentration")
    finished_library_concentration_nm: float = Field(..., alias="Concentration (nM)")
    finished_library_size: Optional[int] = Field(None, alias="Size (bp)")


class EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeFields(BaseStep):
    process_udfs: EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeProcessUDFS
    artifact_udfs: EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeArtifactUDF

    class Config:
        allow_population_by_field_name = True


def get_end_repair(
    lims: Lims, sample_id: str, prep_id: str
) -> EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeFields:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free DNA)",
    )

    return EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeFields(
        **analyte.base_fields(),
        process_udfs=EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeProcessUDFS(
            **analyte.process_udfs()
        ),
        artifact_udfs=EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeArtifactUDF(
            **analyte.artifact_udfs()
        ),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="end_repair",
        workflow="WGS",
    )



#!/usr/bin/env python
import logging
import sys

import click
from genologics.entities import Process
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts


def udf_copy_artifact_to_sample(
    artifacts: list, sample_udf: str, art_udf: str, sample_qc_udf: str = None
) -> None:
    """Function to copy artifact udf and qc to sample level.

    For each artifact in the artifacts list, copying the art_udf and qc_flagg
    to all samples related to the artifact. If a art is a pool, art.samples
    is a list with many samples. Otherwiese a list with only one sample.

    Arguments:
        artifacts: list of artifacts to copy from
        art_udf: artifact udf to get
        sample_udf: sample udf to set
        sample_qc_udf: sample qc udf to set based on artifact qc_flagg"""

    failed_udfs = 0
    passed_udfs = 0
    for art in artifacts:
        udf = art.udf.get(art_udf)
        if udf is not None:
            for sample in art.samples:
                sample.udf[sample_udf] = udf
                if sample_qc_udf:
                    if art.qc_flag == "PASSED":
                        sample.udf[sample_qc_udf] = "True"
                    else:
                        sample.udf[sample_qc_udf] = "False"
                sample.put()
                passed_udfs += 1
        else:
            failed_udfs += 1

    if failed_udfs:
        raise MissingUDFsError(
            message=f"The udf '{art_udf}' is missing for {failed_udfs} artifacts. Udfs were set on {passed_udfs} samples."
        )


@click.command()
@options.sample_udf(help="Sample udf to set.")
@options.artifact_udf(help="Artifact udf to get.")
@options.sample_qc_udf()
@options.input(
    help="Use this flag if you want copy udfs from input artifacts. Defaulte is output artifacts."
)
@click.pass_context
def artifact_to_sample(ctx, sample_udf, artifact_udf, input, sample_qc_udf):
    """Script to copy artifact udf to sample udf"""

    process = ctx.obj["process"]
    try:
        artifacts = get_artifacts(process=process, input=input)
        udf_copy_artifact_to_sample(artifacts, sample_udf, artifact_udf, sample_qc_udf)
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



import logging
from typing import List, Literal

import click
from genologics.lims import Lims, Process, Sample
import requests
from requests import Response
import json

from cg_lims import options
from cg_lims.exceptions import LimsError
from cg_lims.get.samples import get_process_samples
from cg_lims.models.arnold.prep.base_step import BaseStep
from cg_lims.models.arnold.prep.microbial_prep import build_microbial_step_documents
from cg_lims.models.arnold.prep.wgs import build_wgs_documents
from cg_lims.models.arnold.prep.sars_cov_2_prep import build_sars_cov_2_documents
from cg_lims.models.arnold.prep.twist import build_twist_documents

LOG = logging.getLogger(__name__)

prep_document_functions = {
    "wgs": build_wgs_documents,
    "twist": build_twist_documents,
    "micro": build_microbial_step_documents,
    "cov": build_sars_cov_2_documents,
}


def build_step_documents(
    prep_type: Literal["wgs", "twist", "micro", "cov"], process: Process, lims: Lims
) -> List[BaseStep]:
    prep_document_function = prep_document_functions[prep_type]
    samples: List[Sample] = get_process_samples(process=process)
    all_step_documents = []
    for sample in samples:
        step_documents: List[BaseStep] = prep_document_function(
            sample_id=sample.id, process_id=process.id, lims=lims
        )
        all_step_documents += step_documents
    return all_step_documents


@click.command()
@options.prep(help="Prep type.")
@click.pass_context
def prep(ctx, prep_type: Literal["wgs", "twist", "micro", "cov"]):
    """Creating Step documents from a prep in the arnold step collection."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process: Process = ctx.obj["process"]
    lims: Lims = ctx.obj["lims"]
    arnold_host: str = ctx.obj["arnold_host"]

    all_step_documents: List[BaseStep] = build_step_documents(
        prep_type=prep_type, process=process, lims=lims
    )

    response: Response = requests.post(
        url=f"{arnold_host}/steps",
        headers={"Content-Type": "application/json"},
        data=json.dumps([doc.dict(exclude_none=True) for doc in all_step_documents]),
    )
    if not response.ok:
        LOG.info(response.text)
        raise LimsError(response.text)

    LOG.info("Arnold output: %s", response.text)
    click.echo("Step documents inserted to arnold database")



import click

from cg_lims.EPPs.udf.set.set_samples_reads_missing import set_reads_missing_on_new_samples


@click.group(invoke_without_command=True)
@click.pass_context
def set(context: click.Context):
    """Main entry point of set commands"""
    pass


set.add_command(set_reads_missing_on_new_samples)



import csv
import string
from pathlib import Path
from typing import List
import pandas as pd

from cg_lims.exceptions import CSVColumnError


def make_plate_file(
    file: str,
    rows: dict,
    headers: list,
    col_range: int = 13,
    row_range: int = 8,
    newline="\n",
    delimiter=",",
):
    """Creating a plate file.

    Arguments:
        file: file path
        headers: Header list.
        rows: dict of rows
            keys: plate position: eg A1, B1, E4, H2...
            values: list of file rows. Same length as the headers list."""

    with open(file, "w", newline=newline) as hamilton_csv:
        wr = csv.writer(hamilton_csv, delimiter=delimiter)
        wr.writerow(headers)
        for col in range(1, col_range):
            for row in string.ascii_uppercase[0:row_range]:
                row = rows.get(f"{row}{col}")
                if row:
                    wr.writerow(row)


def build_csv(rows: List[List[str]], file: Path, headers: List[str]) -> Path:
    """Build csv."""

    with open(file.absolute(), "w", newline="\n") as new_csv:
        wr = csv.writer(new_csv, delimiter=",")
        wr.writerow(headers)
        wr.writerows(rows)

    return file


def sort_csv(file: Path, columns: List[str], well_columns: List[str] = []):
    """This function sorts a csv file based on the list of columns given.

    columns: list of columns to sort on.
        - eg: [ "Destination Container", "Sample Well" ]

    well_columns: Optional. Sub set of columns!
        - eg: ["Sample Well"]
        - Assumed well format like A1, B1 etc. Not A:1, B:1!
        - Will first be sorted numerically by second field. then alphabetically on the first field:
          (A1, B1, C1, A2, B2, C2), not (A1, A2, B1, B2, C1, C2).
    """

    csv_data_frame = pd.read_csv(file.absolute(), delimiter=",")
    temp_sort_columns = []
    sort_columns = []  # columns to sort on

    if not set(well_columns).issubset(set(columns)):
        raise CSVColumnError(message="well_columns must be subset of columns")

    for column in columns:
        if column not in csv_data_frame.columns:
            raise CSVColumnError(message=f"Column {column} is not in the csv file.")
        if column not in well_columns:
            # If the column is not a well_column, just append it to sort_columns
            sort_columns.append(column)
            continue
        # Splitting the well_column into two temporary sort columns that are added to the data frame and to sort_columns
        well_row = f"{column}_row"
        well_col = f"{column}_col"
        csv_data_frame[well_row] = csv_data_frame[column].transform(lambda x: x[0])
        csv_data_frame[well_col] = csv_data_frame[column].transform(lambda x: int(x[1:]))
        sort_columns += [well_col, well_row]
        temp_sort_columns += [well_col, well_row]

    # Now the data frame has two extra sort columns for each well column

    # sorting by sort_columns
    csv_data_frame.sort_values(by=sort_columns, inplace=True)

    # dropping the temp_sort_columns
    sorted_data = csv_data_frame.drop(temp_sort_columns, axis=1)

    sorted_data.to_csv(file.absolute(), index=False)



#!/usr/bin/env python
import logging
import sys

import click
from genologics.entities import Process
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.samples import get_process_samples



@click.command()
@options.sample_udf(help="Sample udf to set.")
@options.process_udf(help="Process udf to get.")
@click.pass_context
def process_to_sample(ctx, sample_udf, process_udf):
    """Script to copy Process udf to sample udf"""

    process = ctx.obj["process"]
    failed_samples = []

    udf = process.udf.get(process_udf)
    if not udf:
        sys.exit(f"Process udf: {process_udf} not set!")

    samples = get_process_samples(process=process)
    for sample in samples:
        try:
            sample.udf[sample_udf] = udf
            sample.put()
        except:
            failed_samples.append(sample)

    if not failed_samples:
        click.echo(f"Udf {sample_udf} have been set on all samples.")
    else:
        sys.exit(f"Could not set udfs on {len(failed_samples)} samples.")




from genologics.lims import Lims
from pydantic import Field, BaseModel
from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class BeadPurificationArtifactUDFs(BaseModel):
    bead_purification_size_bp: int = Field(..., alias="Size (bp)")
    bead_purification_concentration: float = Field(..., alias="Concentration")


class BeadPurificationProcessUDFs(BaseModel):
    lot_nr_enrichment_kit_hybridization: str = Field(..., alias="Twist Enrichment Kit")
    lot_nr_etoh_bead_purification_post_hyb: str = Field(..., alias="Ethanol")
    lot_nr_h2o_bead_purification_post_hyb: str = Field(..., alias="Nuclease free water")
    bead_purification_post_hyb_method: str = Field(..., alias="Method document")


class BeadPurificationFields(BaseStep):
    artifact_udfs: BeadPurificationArtifactUDFs
    process_udfs: BeadPurificationProcessUDFs

    class Config:
        allow_population_by_field_name = True


def get_bead_purification_twist(lims: Lims, sample_id: str, prep_id: str) -> BeadPurificationFields:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Bead Purification TWIST v2",
    )

    return BeadPurificationFields(
        **analyte.base_fields(),
        process_udfs=BeadPurificationProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=BeadPurificationArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="bead_purification",
        workflow="TWIST",
    )



import click

# commands
from cg_lims.EPPs.files.hamilton.normalization_file import barcode_file
from cg_lims.EPPs.files.hamilton.make_kapa_csv import make_kapa_csv
from cg_lims.EPPs.files.hamilton.sars_cov2_prep_file import sars_cov2_prep_file


@click.group(invoke_without_command=True)
@click.pass_context
def hamilton(ctx):
    """Main entry point of hamilton file commands"""
    pass


hamilton.add_command(make_kapa_csv)
hamilton.add_command(barcode_file)
hamilton.add_command(sars_cov2_prep_file)



from typing import List

from genologics.lims import Lims

from .fragment_dna_truseq_dna import (
    FragmentDNATruSeqDNAFields,
    FragmentDNATruSeqDNAProcessUDFS,
    get_fragemnt_dna_truseq,
)
from .aliquot_sampels_for_covaris import (
    AliquotSamplesforCovarisFields,
    AliquotSamplesforCovarisArtifactUDF,
    get_aliquot_samples_for_covaris,
)
from .reception_control import SampleArtifactUDF, SampleArtifactFields, get_sample_artifact_fields
from .endrepair_size_selection_a_tailing_adapter_ligation import (
    EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeFields,
    EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeProcessUDFS,
    EndRepairSizeSelectionATailingAndAdapterligationTruSeqPCRFreeArtifactUDF,
    get_end_repair,
)
from ..base_step import BaseStep


def build_wgs_documents(sample_id: str, process_id: str, lims: Lims) -> List[BaseStep]:
    """Building a sars_cov_2 Prep."""

    prep_id = f"{sample_id}_{process_id}"
    step_documents = [
        get_fragemnt_dna_truseq(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_sample_artifact_fields(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_aliquot_samples_for_covaris(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_end_repair(sample_id=sample_id, lims=lims, prep_id=prep_id),
    ]
    return [document for document in step_documents if document is not None]



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class FragmentDNATruSeqDNAProcessUDFS(BaseModel):
    lot_nr_covaris_tube: str = Field(..., alias="Lot no: Covaris tube")
    fragmentation_method: str = Field(..., alias="Method Document 1")
    covaris_protocol: str = Field(..., alias="Protocol")
    lot_nr_resuspension_buffer_fragmentation: str = Field(..., alias="Lot no: Resuspension Buffer")


class FragmentDNATruSeqDNAFields(BaseStep):
    process_udfs: FragmentDNATruSeqDNAProcessUDFS

    class Config:
        allow_population_by_field_name = True


def get_fragemnt_dna_truseq(lims: Lims, sample_id: str, prep_id: str) -> FragmentDNATruSeqDNAFields:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Fragment DNA (TruSeq DNA)",
    )

    return FragmentDNATruSeqDNAFields(
        **analyte.base_fields(),
        process_udfs=FragmentDNATruSeqDNAProcessUDFS(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="fragment_dna",
        workflow="WGS",
    )



from typing import Optional
from genologics.lims import Lims
from pydantic import Field, BaseModel

from cg_lims.models.arnold.prep.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class KAPALibraryPreparationArtifactUDFs(BaseModel):
    library_size_pre_hyb: int = Field(..., alias="Size (bp)")
    library_concentration_pre_hyb: float = Field(..., alias="Concentration")
    adapter_ligation_master_mix: str = Field(..., alias="Ligation Master Mix")
    library_preparation_pcr_plate: str = Field(..., alias="PCR Plate")
    amount_of_sample_in_pool_ng: float = Field(
        ..., alias="Amount taken (ng)"
    )  # this value is set afterwards in the pooling step


class KAPALibraryPreparationProcessUDFs(BaseModel):
    hamilton_post_lab_library_preparation: Optional[str] = Field(alias="Hamilton (post-lab)")
    hamilton_pre_lab_library_preparation: Optional[str] = Field(alias="Hamilton (pre-lab)")
    method_document_library_preparation: str = Field(..., alias="Method document")
    library_preparation_kit: str = Field(..., alias="KAPA HyperPlus/Prep Kit")
    lot_nr_h2o_library_preparation_pre_lab: str = Field(..., alias="Nuclease-free water (pre-lab)")
    lot_nr_xgen_adapter: str = Field(..., alias="xGen Adapter")
    lot_nr_index: str = Field(..., alias="Indexing Primer")
    lot_nr_beads_library_preparation_pre_lab: str = Field(
        ..., alias="DNA purification beads (pre-lab)"
    )
    lot_nr_etoh_library_preparation_pre_lab: str = Field(..., alias="Ethanol (pre-lab)")
    pcr_instrument_end_repair_a_tail: str = Field(..., alias="PCR Machine: End Repair and A-tail")
    pcr_instrument_adapter_ligation: str = Field(..., alias="PCR Machine: Adapter ligation")
    pcr_instrument_amplification_plate1: Optional[str] = Field(
        None, alias="PCR Machine: Amplification Plate 1"
    )
    pcr_instrument_amplification_plate2: Optional[str] = Field(
        None, alias="PCR Machine: Amplification Plate 2"
    )
    pcr_instrument_amplification_plate3: Optional[str] = Field(
        None, alias="PCR Machine: Amplification Plate 3"
    )
    lot_nr_beads_library_preparation_post_lab: str = Field(
        ..., alias="DNA purification beads (post-lab)"
    )
    lot_nr_etoh_library_preparation_post_lab: str = Field(..., alias="Ethanol (post-lab)")
    lot_nr_h2o_library_preparation_post_lab: str = Field(
        ..., alias="Nuclease-free water (post-lab)"
    )


class KAPALibraryPreparationFields(BaseStep):
    process_udfs: KAPALibraryPreparationProcessUDFs
    artifact_udfs: KAPALibraryPreparationArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_kapa_library_preparation_twist(
    lims: Lims, sample_id: str, prep_id: str
) -> KAPALibraryPreparationFields:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="KAPA Library Preparation TWIST v1",
    )

    return KAPALibraryPreparationFields(
        **analyte.base_fields(),
        process_udfs=KAPALibraryPreparationProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=KAPALibraryPreparationArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="kapa_library_preparation",
        workflow="TWIST",
    )



import logging
import sys

import click
from requests.exceptions import ConnectionError

from cg_lims.exceptions import LimsError, MissingCgFieldError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def missing_reads(artifacts: list, status_db) -> tuple:
    failed_arts = 0
    passed_arts = 0
    for art in artifacts:
        samples = art.samples
        try:
            reads_total = samples[0].udf["Total Reads (M)"]
            app_tag = samples[0].udf["Sequencing Analysis"]
        except:
            failed_arts += 1
            continue
        try:
            target_amount_reads = status_db.apptag(tag_name=app_tag, key="target_reads")
            guaranteed_fraction = 0.01 * status_db.apptag(
                tag_name=app_tag, key="percent_reads_guaranteed"
            )
        except ConnectionError:
            raise LimsError(message="Could not communicate with cg server")
        except:
            raise MissingCgFieldError(f"Could not find application tag: {app_tag} in database.")
        # Converting from reads to million reads.
        target_amount = target_amount_reads / 1000000
        reads_min = target_amount * guaranteed_fraction
        reads_missing = reads_min - reads_total
        if reads_missing > 0:
            for sample in samples:
                sample.udf["Reads missing (M)"] = target_amount - reads_total
                sample.put()
            art.udf["Rerun"] = True
            art.qc_flag = "FAILED"
        else:
            for sample in samples:
                sample.udf["Reads missing (M)"] = 0
                sample.put()
            art.udf["Rerun"] = False
            art.qc_flag = "PASSED"
        art.put()
        passed_arts += 1
    return passed_arts, failed_arts


@click.command()
@click.pass_context
def get_missing_reads(ctx):
    """Script to calculate missing reads and decide on reruns"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]
    status_db = ctx.obj["status_db"]

    try:
        artifacts = get_artifacts(process=process, input=False)
        passed_arts, failed_arts = missing_reads(artifacts=artifacts, status_db=status_db)
        message = f"Updataed {passed_arts}. Ignored {failed_arts} due to missing UDFs"
        if failed_arts:
            LOG.error(message)
            sys.exit(message)
        else:
            LOG.info(message)
            click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



#!/usr/bin/env python

import click

from cg_lims.EPPs.udf.calculate import calculate

# commands
from cg_lims.EPPs.udf.copy import copy
from cg_lims.EPPs.udf.set import set


@click.group(invoke_without_command=True)
@click.pass_context
def udf(ctx):
    """Main entry point of udf commands"""
    pass


udf.add_command(copy)
udf.add_command(calculate)
udf.add_command(set)



#!/usr/bin/env python
import logging
import pathlib

import click
from genologics.entities import Process

from cg_lims import options
from cg_lims.EPPs.files import files

# commands
from cg_lims.EPPs.move import move
from cg_lims.EPPs.udf import udf
from cg_lims.EPPs.qc import qc
from cg_lims.EPPs.arnold import arnold_upload
from cg_lims.get.files import get_log_content


@click.group(invoke_without_command=True)
@options.log()
@options.process()
@click.pass_context
def epps(ctx, log: str, process: Process):
    """Main entry point of epp commands"""

    log_path = pathlib.Path(log)

    log_content = get_log_content(ctx.obj["lims"], log)
    if log_content:
        with open(log_path.absolute(), "a") as new_log:
            new_log.write(log_content)

    logging.basicConfig(filename=str(log_path.absolute()), filemode="a", level=logging.INFO)
    process = Process(ctx.obj["lims"], id=process)
    ctx.obj["process"] = process


epps.add_command(move)
epps.add_command(files)
epps.add_command(udf)
epps.add_command(qc)
epps.add_command(arnold_upload)



import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


# The maximum amount taken into the prep is MAXIMUM_SAMPLE_AMOUNT.
# Any amount below this can be used in the prep if the total amount is limited.
MAXIMUM_SAMPLE_AMOUNT = 250


def set_amount_needed(artifacts: List[Artifact]):
    """The maximum amount taken into the prep is MAXIMUM_SAMPLE_AMOUNT.
    Any amount below this can be used in the prep if the total amount is limited."""

    missing_udfs = 0
    for art in artifacts:
        amount = art.udf.get("Amount (ng)")
        if not amount:
            missing_udfs += 1
            continue
        if amount >= MAXIMUM_SAMPLE_AMOUNT:
            art.udf["Amount needed (ng)"] = MAXIMUM_SAMPLE_AMOUNT
        else:
            art.udf["Amount needed (ng)"] = amount
        art.put()

    if missing_udfs:
        raise MissingUDFsError(f"Udf missing for {missing_udfs} samples")


@click.command()
@click.pass_context
def twist_aliquot_amount(ctx):
    """Calculates amount needed for samples."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]

    try:
        artifacts = get_artifacts(process=process, input=False)
        set_amount_needed(artifacts)
        message = "Amount needed has been calculated for all samples."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



from pydantic import BaseModel, Field, validator
import datetime as dt
from typing import Optional
import logging

LOG = logging.getLogger(__name__)


def get_number_of_days(first_date: dt, second_date: dt) -> Optional[int]:
    """Get number of days between different time stamps."""

    days = None
    if first_date and second_date:
        time_span = second_date - first_date
        days = time_span.days

    return days


class Sample(BaseModel):
    id: str
    name: str
    project: str
    received_date: Optional[dt.date] = Field(None, alias="Received at")
    prepared_date: Optional[dt.date] = Field(None, alias="Library Prep Finished")
    sequenced_date: Optional[dt.date] = Field(None, alias="Sequencing Finished")
    delivery_date: Optional[dt.date] = Field(None, alias="Delivered at")
    application: Optional[str] = Field(None, alias="Sequencing Analysis")
    bait_set: Optional[str] = Field(None, alias="Bait Set")
    capture_kit: Optional[str] = Field(None, alias="Capture Library version")
    collection_date: Optional[dt.date] = Field(None, alias="Collection Date")
    comment: Optional[str] = Field(None, alias="Comment")
    concentration: Optional[float] = Field(None, alias="Concentration (nM)")
    concentration_sample: Optional[float] = Field(None, alias="Sample Conc.")
    customer: Optional[str] = Field(None, alias="customer")
    data_analysis: Optional[str] = Field(None, alias="Data Analysis")
    data_delivery: Optional[str] = Field(None, alias="Data Delivery")
    elution_buffer: Optional[str] = Field(None, alias="Sample Buffer")
    extraction_method: Optional[str] = Field(None, alias="Extraction method")
    family_name: Optional[str] = Field(None, alias="familyID")
    formalin_fixation_time: Optional[str] = Field(None, alias="Formalin Fixation Time")
    index: Optional[str] = Field(None, alias="Index type")
    index_number: Optional[str] = Field(None, alias="Index number")
    lab_code: Optional[str] = Field(None, alias="Lab Code")
    organism: Optional[str] = Field(None, alias="Strain")
    organism_other: Optional[str] = Field(None, alias="Other species")
    original_lab: Optional[str] = Field(None, alias="Original Lab")
    original_lab_address: Optional[str] = Field(None, alias="Original Lab Address")
    pool: Optional[str] = Field(None, alias="pool name")
    post_formalin_fixation_time: Optional[str] = Field(None, alias="Post Formalin Fixation Time")
    pre_processing_method: Optional[str] = Field(None, alias="Pre Processing Method")
    priority: Optional[str] = Field(None, alias="priority")
    quantity: Optional[str] = Field(None, alias="Quantity")
    reference_genome: Optional[str] = Field(None, alias="Reference Genome Microbial")
    region: Optional[str] = Field(None, alias="Region")
    region_code: Optional[str] = Field(None, alias="Region Code")
    require_qcok: Optional[str] = Field(None, alias="Process only if QC OK")
    rml_plate_name: Optional[str] = Field(None, alias="RML plate name")
    selection_criteria: Optional[str] = Field(None, alias="Selection Criteria")
    sequencing_qc_pass: Optional[str] = Field(None, alias="Passed Sequencing QC")
    sex: Optional[str] = Field(None, alias="Gender")
    source: Optional[str] = Field(None, alias="Source")
    target_reads: Optional[str] = Field(None, alias="Reads missing (M)")
    tissue_block_size: Optional[str] = Field(None, alias="Tissue Block Size")
    tumour: Optional[str] = Field(None, alias="tumor")
    tumour_purity: Optional[str] = Field(None, alias="tumour purity")
    verified_organism: Optional[str] = Field(None, alias="Verified organism")
    volume: Optional[float] = Field(None, alias="Volume (uL)")
    well_position_rml: Optional[str] = Field(None, alias="RML well position")
    processing_time: Optional[int]
    sequenced_to_delivered: Optional[int]
    prepped_to_sequenced: Optional[int]
    received_to_prepped: Optional[int]

    @validator("processing_time", always=True)
    def set_processing_time(cls, v, values: dict) -> Optional[int]:
        """Calculating sample processing time."""
        received_date: Optional[dt.date] = values.get("received_date")
        delivery_date: Optional[dt.date] = values.get("delivery_date")
        return get_number_of_days(first_date=received_date, second_date=delivery_date)

    @validator("sequenced_to_delivered", always=True)
    def set_sequenced_to_delivered(cls, v, values: dict) -> Optional[int]:
        """Calculating sample processing time."""
        received_date: Optional[dt.date] = values.get("sequenced_date")
        delivery_date: Optional[dt.date] = values.get("delivered_date")
        return get_number_of_days(first_date=received_date, second_date=delivery_date)

    @validator("prepped_to_sequenced", always=True)
    def set_prepped_to_sequenced(cls, v, values: dict) -> Optional[int]:
        """Calculating sample processing time."""
        received_date: Optional[dt.date] = values.get("prepared_date")
        delivery_date: Optional[dt.date] = values.get("sequenced_date")
        return get_number_of_days(first_date=received_date, second_date=delivery_date)

    @validator("received_to_prepped", always=True)
    def set_received_to_prepped(cls, v, values: dict) -> Optional[int]:
        """Calculating sample processing time."""
        received_date: Optional[dt.date] = values.get("received_date")
        delivery_date: Optional[dt.date] = values.get("prepared_date")
        return get_number_of_days(first_date=received_date, second_date=delivery_date)

    class Config:
        allow_population_by_field_name = True



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class LibraryPreparationCovProcessUDFS(BaseModel):
    """Library Preparation (Cov) v1"""

    lot_nr_tagmentation_beads: str = Field(..., alias="Tagmentation beads")
    lot_nr__stop_tagment_buffer: str = Field(..., alias="Stop Tagment Buffer")
    lot_nr_index: str = Field(..., alias="Index")
    lot_nr_pcr_mix: str = Field(..., alias="PCR-mix")
    lot_nr_tagmentation_wash_buffer: str = Field(..., alias="Tagmentation Wash Buffer")
    lot_nr_h2o_library_preparation: str = Field(..., alias="Nuclease-free water")
    lot_nr_TB1: str = Field(..., alias="TB1 HT")
    pcr_instrument_tagmentation: str = Field(..., alias="PCR machine: Tagmentation")
    pcr_instrument_amplification: str = Field(..., alias="PCR machine: Amplification")
    library_preparation_method: str = Field(..., alias="Method document")
    liquid_handling_system: str = Field(..., alias="Instrument")


class LibraryPreparationCovFields(
    BaseStep,
):
    process_udfs: LibraryPreparationCovProcessUDFS

    class Config:
        allow_population_by_field_name = True


def get_library_prep_cov(lims: Lims, sample_id: str, prep_id: str) -> LibraryPreparationCovFields:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Library Preparation (Cov) v1",
    )

    return LibraryPreparationCovFields(
        **analyte.base_fields(),
        process_udfs=LibraryPreparationCovProcessUDFS(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="library_prep",
        workflow="sars_cov_2"
    )



#!/usr/bin/env python

import click

from cg_lims.EPPs.arnold.prep import prep
from cg_lims.EPPs.arnold.sample import sample


@click.group(invoke_without_command=True)
@click.pass_context
def arnold_upload(ctx):
    """Main load commands."""
    pass


arnold_upload.add_command(prep)
arnold_upload.add_command(sample)



import logging
import sys
from typing import List

import click
from genologics.entities import Artifact, Process

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def calculate_volumes(artifacts: List[Artifact], process: Process):
    """Calculates volumes of H2O for diluting samples. The total volume differ depending on
    type of sample. It is given by the process udf 'Total Volume (ul)'."""

    max_volume = process.udf.get("Total Volume (ul)")
    if not max_volume:
        raise MissingUDFsError("Process udf missing: Total Volume (ul)")

    missing_udfs = 0

    for artifact in artifacts:
        sample_volume = artifact.udf.get("Volume (ul)")

        if sample_volume is None:
            missing_udfs += 1
            continue
        volume_water = max_volume - sample_volume
        if volume_water < 0:
            volume_water = 0
        artifact.udf["Volume H2O (ul)"] = volume_water
        artifact.put()

    if missing_udfs:
        raise MissingUDFsError(f"Udf missing for {missing_udfs} samples")


@click.command()
@click.pass_context
def volume_water(ctx):
    """Calculates water volume for samples."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]

    try:
        artifacts = get_artifacts(process=process, input=False)
        calculate_volumes(artifacts=artifacts, process=process)
        message = "Volume H2O have been calculated for all samples."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



#!/usr/bin/env python
from typing import Tuple

from genologics.entities import Process, Artifact
from genologics.lims import Lims

from cg_lims.exceptions import LimsError, MissingUDFsError

from cg_lims.get.artifacts import get_artifacts, get_qc_output_artifacts
from cg_lims.get.udfs import get_udf_type
from cg_lims import options

import logging
import sys

import click

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def find_reads_to_sequence(process: Process, lims: Lims) -> Tuple[int, int]:
    """Function to copy Missing reads Pool (M) from input artifact if existing,
     otherwise Reads missing (M) from sample."""
    
    failed_artifacts = 0
    passed_artifacts = 0

    for input, output in process.input_output_maps:
        if output.get('output-type') != 'Analyte':
            continue

        input_artifact = Artifact(lims, id=input['limsid'])
        output_artifact = Artifact(lims, id=output['limsid'])
        if input_artifact.udf.get('Missing reads Pool (M)') is not None:
            reads_to_sequence = input_artifact.udf.get('Missing reads Pool (M)')
        elif output_artifact.samples[0].udf.get('Reads missing (M)') is not None:
            reads_to_sequence = output_artifact.samples[0].udf.get('Reads missing (M)')
        else:
            failed_artifacts +=1
            continue

        output_artifact.udf['Reads to sequence (M)'] = str(reads_to_sequence)
        output_artifact.put()
        passed_artifacts += 1
    return passed_artifacts, failed_artifacts



@click.command()
@click.pass_context
def reads_to_sequence(ctx):
    """Script to find Reads to sequence (M) for a sample or a pool"""

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    try:
        find_reads_to_sequence(process=process, lims=lims)
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



#!/usr/bin/env python

import logging
import sys
import csv
from pathlib import Path

import click
from genologics.entities import Artifact

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingArtifactError, MissingFileError
from cg_lims.get.artifacts import get_artifact_by_name
from cg_lims.get.files import get_file_path

LOG = logging.getLogger(__name__)


def make_well_dict(process, lims, input):
    """Creates a well dict based on input_output_map
    keys: well of input artifact
    values: input/output artifact depending on the input flag
    """

    well_dict = {}
    for inp, outp in process.input_output_maps:
        if outp.get("output-generation-type") == "PerAllInputs":
            continue
        in_art = Artifact(lims, id=inp["limsid"])
        out_art = Artifact(lims, id=outp["limsid"])
        source_art = in_art if input == True else out_art
        col, row = source_art.location[1].split(":")
        well = col + row
        well_dict[well] = out_art
    return well_dict


def set_udfs(well_field: str, value_field: str, udf: str, well_dict: dict, result_file: Path):
    """Reads the csv and sets the value for each sample"""

    error_msg = []
    passed_arts = 0
    with open(result_file, newline="", encoding="latin1") as csvfile:
        reader = csv.DictReader(csvfile)
        for sample in reader:
            well = sample.get(well_field)
            value = sample.get(value_field)
            if value is None:
                error_msg.append("Some samples in the file hade missing values.")
                continue
            elif well not in well_dict:
                error_msg.append("Some samples in the step were not represented in the file.")
                continue
            art = well_dict[well]
            try:
            	art.udf[udf] = str(value)
            except:
                art.udf[udf] = float(value)
            art.put()
            passed_arts += 1

    if passed_arts < len(well_dict.keys()):
        error_msg.append("Some samples in the step were not represented in the file.")

    error_string = ' '.join(list(set(error_msg)))
    if error_msg:
        raise MissingArtifactError(error_string)


@click.command()
@options.file_placeholder(help="File placeholder name.")
@options.local_file()
@options.udf()
@options.well_field()
@options.value_field()
@options.input()
@click.pass_context
def csv_well_to_udf(
    ctx, file: str, well_field: str, value_field: str, udf: str, input: bool, local_file: str
):
    """Script to copy data from file to udf based on well position"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    if local_file:
        file_path = local_file
    else:
        file_art = get_artifact_by_name(process=process, name=file)
        file_path = get_file_path(file_art)

    try:
        if not Path(file_path).is_file():
            raise MissingFileError(f"No such file: {file_path}")
        well_dict = make_well_dict(process, lims, input)
        set_udfs(well_field, value_field, udf, well_dict, file_path)
        click.echo("The udfs were sucessfully populated.")
    except LimsError as e:
        sys.exit(e.message)



from typing import Optional
from genologics.lims import Lims
from pydantic import Field, BaseModel
from cg_lims.objects import BaseAnalyte

from cg_lims.models.arnold.prep.base_step import BaseStep


class PoolsamplesforhybridizationArtifactUDFs(BaseModel):
    amount_of_sample: Optional[float] = Field(None, alias="Total Amount (ng)")


class PoolsamplesforhybridizationProcessUDFs(BaseModel):

    speedvac_temp: float = Field(..., alias="SpeedVac temp")
    speedvac_time: float = Field(..., alias="SpeedVac time (min)")


class PoolSamplesForHybridizationFields(BaseStep):
    process_udfs: PoolsamplesforhybridizationProcessUDFs
    artifact_udfs: PoolsamplesforhybridizationArtifactUDFs
    #!!! same as container_name??? https://clinical-lims.scilifelab.se/clarity/work-complete/159432  Optional[str] = Field(None, alias="container_name")

    class Config:
        allow_population_by_field_name = True


def get_pool_samples_twist(
    lims: Lims, sample_id: str, prep_id: str
) -> PoolSamplesForHybridizationFields:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="pool samples TWIST v2",
    )

    return PoolSamplesForHybridizationFields(
        **analyte.base_fields(),
        process_udfs=PoolsamplesforhybridizationProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=PoolsamplesforhybridizationArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="pool_samples",
        workflow="TWIST",
    )



import click
from genologics.entities import Process


@click.command()
@click.pass_context
def set_qc_fail(ctx):
    """Helper script for procudtion so set qc-flaggs to failed after the step has been closed.

    The script will set qc flagg to failed on all artifacts within a process,
     where the process is defined by the udf "QC Process ID".
     The script is supposed to be run from within a script process."""

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    qc_process_id = process.udf.get("QC Process ID")
    qc_process = Process(lims=lims, id=qc_process_id)
    artifacts = qc_process.all_outputs(unique=True)
    for artifact in artifacts:
        artifact.qc_flag = "FAILED"
        artifact.put()
    click.echo("QC-flags have been set.")
    process.udf["QC Process ID"] = ""
    process.put()
